@@ -47,6 +47,53 @@
  * A plugin for operations with file systems
  */
 
+typedef enum {
+    BD_FS_RESIZE,
+    BD_FS_REPAIR,
+    BD_FS_CHECK,
+    BD_FS_LABEL
+} BDFsOpType;
+
+/**
+ * BDFSInfo:
+ * @type: filesystem identifier, must be present
+ * @check_util: required utility for consistency checking, "" if not needed and NULL for no support
+ * @repair_util: required utility for repair, "" if not needed and NULL for no support
+ * @resize_util: required utility for resize, "" if not needed and NULL for no support
+ * @resize_mode: resize availability flags, 0 if no support
+ */
+typedef struct BDFSInfo
+{
+    const gchar *type;
+    const gchar *check_util;
+    const gchar *repair_util;
+    const gchar *resize_util;
+    BDFsResizeFlags resize_mode;
+    const gchar *label_util;
+} BDFSInfo;
+
+const BDFSInfo fs_info[] = {
+    {"xfs", "xfs_db", "xfs_repair", "xfs_growfs", BD_FS_ONLINE_GROW | BD_FS_OFFLINE_GROW, "xfs_admin"},
+    {"ext2", "e2fsck", "e2fsck", "resize2fs", BD_FS_ONLINE_GROW | BD_FS_OFFLINE_GROW | BD_FS_OFFLINE_SHRINK, "tune2fs"},
+    {"ext3", "e2fsck", "e2fsck", "resize2fs", BD_FS_ONLINE_GROW | BD_FS_OFFLINE_GROW | BD_FS_OFFLINE_SHRINK, "tune2fs"},
+    {"ext4", "e2fsck", "e2fsck", "resize2fs", BD_FS_ONLINE_GROW | BD_FS_OFFLINE_GROW | BD_FS_OFFLINE_SHRINK, "tune2fs"},
+    {"vfat", "fsck.vfat", "fsck.vfat", "", BD_FS_OFFLINE_GROW | BD_FS_OFFLINE_SHRINK, "fatlabel"},
+    {NULL, NULL, NULL, NULL, 0, NULL}
+};
+
+static const BDFSInfo *
+get_fs_info (const gchar *type)
+{
+    g_return_val_if_fail (type != NULL, NULL);
+
+    for (guint n = 0; fs_info[n].type != NULL; n++) {
+        if (strcmp (fs_info[n].type, type) == 0)
+            return &fs_info[n];
+    }
+
+    return NULL;
+}
+
 /**
  * bd_fs_error_quark: (skip)
  */
@@ -197,101 +244,7 @@ typedef gboolean (*MountFunc) (MountArgs *args, GError **error);
  *
  */
 gboolean bd_fs_check_deps () {
-    GError *error = NULL;
-    gboolean check_ret = TRUE;
-    gboolean ret = bd_utils_check_util_version ("mke2fs", NULL, "", NULL, &error);
-
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("e2fsck", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("tune2fs", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("dumpe2fs", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("resize2fs", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("mkfs.xfs", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("xfs_db", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("xfs_repair", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("xfs_admin", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("xfs_growfs", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("mkfs.vfat", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("fatlabel", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    ret = bd_utils_check_util_version ("fsck.vfat", NULL, "", NULL, &error);
-    if (!ret && error) {
-        g_warning("Cannot load the FS plugin: %s" , error->message);
-        g_clear_error (&error);
-        check_ret = FALSE;
-    }
-
-    return check_ret;
+    return TRUE;
 }
 
 /**
@@ -476,6 +429,12 @@ static gboolean do_mount (MountArgs *args, GError **error) {
         }
     }
 
+#ifdef LIBMOUNT_RO_FALLBACK
+    /* we don't want libmount to try RDONLY mounts if we were explicitly given the "rw" option */
+    if (args->options && (mnt_optstr_get_option (args->options, "rw", NULL, NULL) == 0))
+        mnt_context_enable_rwonly_mount (cxt, TRUE);
+#endif
+
     ret = mnt_context_mount (cxt);
     if (ret != 0) {
         if (mnt_context_get_mflags (cxt, &mflags) != 0) {
@@ -524,11 +483,12 @@ static gboolean do_mount (MountArgs *args, GError **error) {
                                      "Filesystem type %s not configured in kernel.", args->fstype);
                     break;
                 case EROFS:
+                case EACCES:
                       if (mflags & MS_RDONLY) {
                           g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_FAIL,
                                        "Cannot mount %s read-only.", args->device);
                           break;
-                      } else if (mnt_optstr_get_option (args->options, "rw", NULL, NULL) == 0) {
+                      } else if (args->options && (mnt_optstr_get_option (args->options, "rw", NULL, NULL) == 0)) {
                           g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_FAIL,
                                        "%s is write-protected but `rw' option given.", args->device);
                           break;
@@ -536,7 +496,10 @@ static gboolean do_mount (MountArgs *args, GError **error) {
                           g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_FAIL,
                                        "Mount %s on %s failed.", args->device, args->mountpoint);
                           break;
-                      } else {
+                      }
+#ifndef LIBMOUNT_RO_FALLBACK
+                      /* new versions of libmount do this automatically */
+                      else {
                           MountArgs ro_args;
                           gboolean success = FALSE;
 
@@ -555,7 +518,9 @@ static gboolean do_mount (MountArgs *args, GError **error) {
 
                           return success;
                       }
-
+#else
+                      break;
+#endif
                 default:
                     g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_FAIL,
                                  "Mount syscall failed: %d.", syscall_errno);
@@ -1383,12 +1348,27 @@ static gboolean ext_mkfs (const gchar *device, const BDExtraArg **extra, const g
     return bd_utils_exec_and_report_error (args, extra, error);
 }
 
+/**
+ * xfs_resize_device:
+ * @device: the device the file system of which to resize
+ * @new_size: new requested size for the file system *in bytes*
+ *            (if 0, the file system is adapted to the underlying block device)
+ * @extra: (allow-none) (array zero-terminated=1): extra options for the resize (right now
+ *                                                 passed to the 'xfs_growfs' utility)
+ * @error: (out): place to store error (if any)
+ *
+ * This is just a helper function for bd_fs_resize.
+ *
+ * Returns: whether the file system on @device was successfully resized or not
+ */
+
 static gboolean xfs_resize_device (const gchar *device, guint64 new_size, const BDExtraArg **extra, GError **error) {
     g_autofree gchar* mountpoint = NULL;
     gboolean ret = FALSE;
     gboolean success = FALSE;
     gboolean unmount = FALSE;
     GError *local_error = NULL;
+    BDFSXfsInfo* xfs_info = NULL;
 
     mountpoint = bd_fs_get_mountpoint (device, error);
     if (!mountpoint) {
@@ -1414,6 +1394,14 @@ static gboolean xfs_resize_device (const gchar *device, guint64 new_size, const
         }
     }
 
+    xfs_info = bd_fs_xfs_get_info (device, error);
+    if (!xfs_info) {
+        return FALSE;
+    }
+
+    new_size = (new_size + xfs_info->block_size - 1) / xfs_info->block_size;
+    bd_fs_xfs_info_free (xfs_info);
+
     success = bd_fs_xfs_resize (mountpoint, new_size, extra, error);
 
     if (unmount) {
@@ -1437,20 +1425,8 @@ static gboolean xfs_resize_device (const gchar *device, guint64 new_size, const
     return success;
 }
 
-/**
- * bd_fs_resize:
- * @device: the device the file system of which to resize
- * @new_size: new requested size for the file system (if 0, the file system is
- *            adapted to the underlying block device)
- * @error: (out): place to store error (if any)
- *
- * Resize filesystem on @device. This calls other fs resize functions from this
- * plugin based on detected filesystem (e.g. bd_fs_xfs_resize for XFS). This
- * function will return an error for unknown/unsupported filesystems.
- *
- * Returns: whether the file system on @device was successfully resized or not
- */
-gboolean bd_fs_resize (const gchar *device, guint64 new_size, GError **error) {
+static gboolean device_operation (const gchar *device, BDFsOpType op, guint64 new_size, const gchar *label, GError **error) {
+    const gchar* op_name = NULL;
     g_autofree gchar* fstype = NULL;
 
     fstype = bd_fs_get_fstype (device, error);
@@ -1465,22 +1441,241 @@ gboolean bd_fs_resize (const gchar *device, guint64 new_size, GError **error) {
         }
     }
 
-    if (g_strcmp0 (fstype, "ext2") == 0) {
-        return bd_fs_ext2_resize (device, new_size, NULL, error);
-    } else if (g_strcmp0 (fstype, "ext3") == 0) {
-        return bd_fs_ext3_resize (device, new_size, NULL, error);
-    } else if (g_strcmp0 (fstype, "ext4") == 0) {
-        return bd_fs_ext4_resize (device, new_size, NULL, error);
+    if (g_strcmp0 (fstype, "ext2") == 0 || g_strcmp0 (fstype, "ext3") == 0
+                                        || g_strcmp0 (fstype, "ext4") == 0) {
+        switch (op) {
+            case BD_FS_RESIZE:
+                return bd_fs_ext4_resize (device, new_size, NULL, error);
+            case BD_FS_REPAIR:
+                return bd_fs_ext4_repair (device, TRUE, NULL, error);
+            case BD_FS_CHECK:
+                return bd_fs_ext4_check (device, NULL, error);
+            case BD_FS_LABEL:
+                return bd_fs_ext4_set_label (device, label, error);
+        }
     } else if (g_strcmp0 (fstype, "xfs") == 0) {
-        return xfs_resize_device (device, new_size, NULL, error);
+        switch (op) {
+            case BD_FS_RESIZE:
+                return xfs_resize_device (device, new_size, NULL, error);
+            case BD_FS_REPAIR:
+                return bd_fs_xfs_repair (device, NULL, error);
+            case BD_FS_CHECK:
+                return bd_fs_xfs_check (device, error);
+            case BD_FS_LABEL:
+                return bd_fs_xfs_set_label (device, label, error);
+        }
     } else if (g_strcmp0 (fstype, "vfat") == 0) {
-        return bd_fs_vfat_resize (device, new_size,  error);
-    } else {
+        switch (op) {
+            case BD_FS_RESIZE:
+                return bd_fs_vfat_resize (device, new_size, error);
+            case BD_FS_REPAIR:
+                return bd_fs_vfat_repair (device, NULL, error);
+            case BD_FS_CHECK:
+                return bd_fs_vfat_check (device, NULL, error);
+            case BD_FS_LABEL:
+                return bd_fs_vfat_set_label (device, label, error);
+        }
+    }
+
+    switch (op) {
+        case BD_FS_RESIZE:
+            op_name = "Resizing";
+            break;
+        case BD_FS_REPAIR:
+            op_name = "Repairing";
+            break;
+        case BD_FS_CHECK:
+            op_name = "Checking";
+            break;
+        case BD_FS_LABEL:
+            op_name = "Setting the label of";
+            break;
+    }
+    g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_NOT_SUPPORTED,
+                 "%s filesystem '%s' is not supported.", op_name, fstype);
+    return FALSE;
+}
+
+/**
+ * bd_fs_resize:
+ * @device: the device the file system of which to resize
+ * @new_size: new requested size for the file system (if 0, the file system is
+ *            adapted to the underlying block device)
+ * @error: (out): place to store error (if any)
+ *
+ * Resize filesystem on @device. This calls other fs resize functions from this
+ * plugin based on detected filesystem (e.g. bd_fs_xfs_resize for XFS). This
+ * function will return an error for unknown/unsupported filesystems.
+ *
+ * Returns: whether the file system on @device was successfully resized or not
+ */
+gboolean bd_fs_resize (const gchar *device, guint64 new_size, GError **error) {
+    return device_operation (device, BD_FS_RESIZE, new_size, NULL, error);
+}
+
+/**
+ * bd_fs_repair:
+ * @device: the device the file system of which to repair
+ * @error: (out): place to store error (if any)
+ *
+ * Repair filesystem on @device. This calls other fs repair functions from this
+ * plugin based on detected filesystem (e.g. bd_fs_xfs_repair for XFS). This
+ * function will return an error for unknown/unsupported filesystems.
+ *
+ * Returns: whether the file system on @device was successfully repaired or not
+ */
+gboolean bd_fs_repair (const gchar *device, GError **error) {
+    return device_operation (device, BD_FS_REPAIR, 0, NULL, error);
+ }
+
+/**
+ * bd_fs_check:
+ * @device: the device the file system of which to check
+ * @error: (out): place to store error (if any)
+ *
+ * Check filesystem on @device. This calls other fs check functions from this
+ * plugin based on detected filesystem (e.g. bd_fs_xfs_check for XFS). This
+ * function will return an error for unknown/unsupported filesystems.
+ *
+ * Returns: whether the file system on @device passed the consistency check or not
+ */
+gboolean bd_fs_check (const gchar *device, GError **error) {
+    return device_operation (device, BD_FS_CHECK, 0, NULL, error);
+}
+
+/**
+ * bd_fs_set_label:
+ * @device: the device with file system to set the label for
+ * @error: (out): place to store error (if any)
+ *
+ * Set label for filesystem on @device. This calls other fs label functions from this
+ * plugin based on detected filesystem (e.g. bd_fs_xfs_set_label for XFS). This
+ * function will return an error for unknown/unsupported filesystems.
+ *
+ * Returns: whether the file system on @device was successfully relabled or not
+ */
+gboolean bd_fs_set_label (const gchar *device, const gchar *label, GError **error) {
+    return device_operation (device, BD_FS_LABEL, 0, label, error);
+ }
+
+static gboolean query_fs_operation (const gchar *fs_type, BDFsOpType op, gchar **required_utility, BDFsResizeFlags *mode, GError **error) {
+    gboolean ret;
+    const BDFSInfo *fsinfo = NULL;
+    const gchar* op_name = NULL;
+    const gchar* exec_util = NULL;
+
+    if (required_utility != NULL)
+        *required_utility = NULL;
+
+    if (mode != NULL)
+        *mode = 0;
+
+    fsinfo = get_fs_info (fs_type);
+    if (fsinfo != NULL) {
+        switch (op) {
+            case BD_FS_RESIZE:
+                op_name = "Resizing";
+                exec_util = fsinfo->resize_util;
+                break;
+            case BD_FS_REPAIR:
+                op_name = "Repairing";
+                exec_util = fsinfo->repair_util;
+                break;
+            case BD_FS_CHECK:
+                op_name = "Checking";
+                exec_util = fsinfo->check_util;
+                break;
+            case BD_FS_LABEL:
+                op_name = "Setting the label of";
+                exec_util = fsinfo->label_util;
+                break;
+        }
+    }
+
+    if (fsinfo == NULL || exec_util == NULL) {
         g_set_error (error, BD_FS_ERROR, BD_FS_ERROR_NOT_SUPPORTED,
-                     "Resizing filesystem '%s' is not supported.", fstype);
+                     "%s filesystem '%s' is not supported.", op_name, fs_type);
         return FALSE;
     }
 
+    if (mode != NULL)
+        *mode = fsinfo->resize_mode;
+
+    if (strlen(exec_util) == 0) { /* empty string if no util needed */
+        return TRUE;
+    }
+
+    ret = bd_utils_check_util_version (exec_util, NULL, "", NULL, NULL);
+    if (!ret && required_utility != NULL)
+        *required_utility = g_strdup (exec_util);
+
+    return ret;
+}
+
+/**
+ * bd_fs_can_resize:
+ * @type: the filesystem type to be tested for installed resize support
+ * @mode: (out): flags for allowed resizing (i.e. growing/shrinking support for online/offline)
+ * @required_utility: (out) (transfer full): the utility binary which is required for resizing (if missing i.e. returns FALSE but no error)
+ * @error: (out): place to store error (if any)
+ *
+ * Searches for the required utility to resize the given filesystem and returns whether
+ * it is installed. The mode flags indicate if growing and/or shrinking resize is available if
+ * mounted/unmounted.
+ * Unknown filesystems or filesystems which do not support resizing result in errors.
+ *
+ * Returns: whether filesystem resize is available
+ */
+gboolean bd_fs_can_resize (const gchar *type, BDFsResizeFlags *mode, gchar **required_utility, GError **error) {
+    return query_fs_operation (type, BD_FS_RESIZE, required_utility, mode, error);
+}
+
+/**
+ * bd_fs_can_check:
+ * @type: the filesystem type to be tested for installed consistency check support
+ * @required_utility: (out) (transfer full): the utility binary which is required for checking (if missing i.e. returns FALSE but no error)
+ * @error: (out): place to store error (if any)
+ *
+ * Searches for the required utility to check the given filesystem and returns whether
+ * it is installed.
+ * Unknown filesystems or filesystems which do not support checking result in errors.
+ *
+ * Returns: whether filesystem check is available
+ */
+gboolean bd_fs_can_check (const gchar *type, gchar **required_utility, GError **error) {
+    return query_fs_operation (type, BD_FS_CHECK, required_utility, NULL, error);
+}
+
+/**
+ * bd_fs_can_repair:
+ * @type: the filesystem type to be tested for installed repair support
+ * @required_utility: (out) (transfer full): the utility binary which is required for repairing (if missing i.e. return FALSE but no error)
+ * @error: (out): place to store error (if any)
+ *
+ * Searches for the required utility to repair the given filesystem and returns whether
+ * it is installed.
+ * Unknown filesystems or filesystems which do not support reparing result in errors.
+ *
+ * Returns: whether filesystem repair is available
+ */
+gboolean bd_fs_can_repair (const gchar *type, gchar **required_utility, GError **error) {
+    return query_fs_operation (type, BD_FS_REPAIR, required_utility, NULL, error);
+}
+
+/**
+ * bd_fs_can_set_label:
+ * @type: the filesystem type to be tested for installed label support
+ * @required_utility: (out) (transfer full): the utility binary which is required for relabeling (if missing i.e. return FALSE but no error)
+ * @error: (out): place to store error (if any)
+ *
+ * Searches for the required utility to set the label of the given filesystem and returns whether
+ * it is installed.
+ * Unknown filesystems or filesystems which do not support setting the label result in errors.
+ *
+ * Returns: whether setting filesystem label is available
+ */
+gboolean bd_fs_can_set_label (const gchar *type, gchar **required_utility, GError **error) {
+    return query_fs_operation (type, BD_FS_LABEL, required_utility, NULL, error);
 }
 
 /**
