@@ -5,7 +5,7 @@ import subprocess
 import tempfile
 from contextlib import contextmanager
 import utils
-from utils import run
+from utils import run, create_sparse_tempfile
 import six
 import overrides_hack
 
@@ -844,6 +844,86 @@ class VfatResize(FSTestCase):
         succ = BlockDev.fs_vfat_resize(self.loop_dev, 0)
         self.assertTrue(succ)
 
+class CanResizeRepairCheckLabel(FSTestCase):
+    def test_can_resize(self):
+        """Verify that tooling query works for resize"""
+
+        avail, mode, util = BlockDev.fs_can_resize("ext4")
+        self.assertTrue(avail)
+        self.assertEqual(util, None)
+
+        old_path = os.environ.get("PATH", "")
+        os.environ["PATH"] = ""
+        avail, mode, util = BlockDev.fs_can_resize("ext4")
+        os.environ["PATH"] = old_path
+        self.assertFalse(avail)
+        self.assertEqual(util, "resize2fs")
+        self.assertEqual(mode, BlockDev.FsResizeFlags.ONLINE_GROW |
+                               BlockDev.FsResizeFlags.OFFLINE_GROW |
+                               BlockDev.FsResizeFlags.OFFLINE_SHRINK)
+
+        avail = BlockDev.fs_can_resize("vfat")
+        self.assertTrue(avail)
+
+        with self.assertRaises(GLib.GError):
+            BlockDev.fs_can_resize("nilfs2")
+
+    def test_can_repair(self):
+        """Verify that tooling query works for repair"""
+
+        avail, util = BlockDev.fs_can_repair("xfs")
+        self.assertTrue(avail)
+        self.assertEqual(util, None)
+
+        old_path = os.environ.get("PATH", "")
+        os.environ["PATH"] = ""
+        avail, util = BlockDev.fs_can_repair("xfs")
+        os.environ["PATH"] = old_path
+        self.assertFalse(avail)
+        self.assertEqual(util, "xfs_repair")
+
+        avail = BlockDev.fs_can_repair("vfat")
+        self.assertTrue(avail)
+
+        with self.assertRaises(GLib.GError):
+            BlockDev.fs_can_repair("nilfs2")
+
+    def test_can_check(self):
+        """Verify that tooling query works for consistency check"""
+
+        avail, util = BlockDev.fs_can_check("xfs")
+        self.assertTrue(avail)
+        self.assertEqual(util, None)
+
+        old_path = os.environ.get("PATH", "")
+        os.environ["PATH"] = ""
+        avail, util = BlockDev.fs_can_check("xfs")
+        os.environ["PATH"] = old_path
+        self.assertFalse(avail)
+        self.assertEqual(util, "xfs_db")
+
+        avail = BlockDev.fs_can_check("vfat")
+        self.assertTrue(avail)
+
+        with self.assertRaises(GLib.GError):
+            BlockDev.fs_can_check("nilfs2")
+
+    def test_can_set_label(self):
+        """Verify that tooling query works for setting the label"""
+
+        avail, util = BlockDev.fs_can_set_label("xfs")
+        self.assertTrue(avail)
+        self.assertEqual(util, None)
+
+        old_path = os.environ.get("PATH", "")
+        os.environ["PATH"] = ""
+        avail, util = BlockDev.fs_can_set_label("xfs")
+        os.environ["PATH"] = old_path
+        self.assertFalse(avail)
+        self.assertEqual(util, "xfs_admin")
+
+        with self.assertRaises(GLib.GError):
+            BlockDev.fs_can_set_label("nilfs2")
 
 class MountTest(FSTestCase):
 
@@ -915,6 +995,43 @@ class MountTest(FSTestCase):
         self.assertTrue(succ)
         self.assertFalse(os.path.ismount(tmp))
 
+    def test_mount_ro_device(self):
+        """ Test mounting an FS on a RO device """
+
+        backing_file = create_sparse_tempfile("ro_mount", 50 * 1024**2)
+        self.addCleanup(os.unlink, backing_file)
+        self.assertTrue(BlockDev.fs_xfs_mkfs(backing_file, None))
+
+        succ, dev = BlockDev.loop_setup(backing_file, 0, 0, True, False)
+        self.assertTrue(succ)
+        self.addCleanup(BlockDev.loop_teardown, dev)
+
+        tmp_dir = tempfile.mkdtemp(prefix="libblockdev.", suffix="mount_test")
+        self.addCleanup(os.rmdir, tmp_dir)
+
+        loop_dev = "/dev/" + dev
+        # without any options, the mount should fall back to RO
+        self.assertTrue(BlockDev.fs_mount(loop_dev, tmp_dir, None, None, None))
+        self.addCleanup(umount, dev)
+        self.assertTrue(os.path.ismount(tmp_dir))
+
+        succ = BlockDev.fs_unmount(tmp_dir, False, False, None)
+        self.assertTrue(succ)
+        self.assertFalse(os.path.ismount(tmp_dir))
+
+        # explicit "ro" should work just fine too
+        self.assertTrue(BlockDev.fs_mount(loop_dev, tmp_dir, None, "ro", None))
+        self.assertTrue(os.path.ismount(tmp_dir))
+
+        succ = BlockDev.fs_unmount(tmp_dir, False, False, None)
+        self.assertTrue(succ)
+        self.assertFalse(os.path.ismount(tmp_dir))
+
+        # explicit "rw" should fail
+        with self.assertRaises(GLib.GError):
+            BlockDev.fs_mount(loop_dev, tmp_dir, None, "rw", None)
+        self.assertFalse(os.path.ismount(tmp_dir))
+
     @unittest.skipUnless("JENKINS_HOME" in os.environ, "skipping test that modifies system configuration")
     def test_mount_fstab(self):
         """ Test mounting and unmounting devices in /etc/fstab """
@@ -997,55 +1114,181 @@ class MountTest(FSTestCase):
             BlockDev.fs_unmount(self.loop_dev, run_as_uid=uid, run_as_gid=gid)
         self.assertTrue(os.path.ismount(tmp))
 
-
-class GenericResize(FSTestCase):
-    def _test_generic_resize(self, mkfs_function):
+class GenericCheck(FSTestCase):
+    def _test_generic_check(self, mkfs_function):
         # clean the device
         succ = BlockDev.fs_clean(self.loop_dev)
 
         succ = mkfs_function(self.loop_dev, None)
         self.assertTrue(succ)
 
+        # check for consistency (expected to be ok)
+        succ = BlockDev.fs_check(self.loop_dev)
+        self.assertTrue(succ)
+
+    def test_ext4_generic_check(self):
+        """Test generic check function with an ext4 file system"""
+        self._test_generic_check(mkfs_function=BlockDev.fs_ext4_mkfs)
+
+    def test_xfs_generic_check(self):
+        """Test generic check function with an ext4 file system"""
+        self._test_generic_check(mkfs_function=BlockDev.fs_xfs_mkfs)
+
+class GenericRepair(FSTestCase):
+    def _test_generic_repair(self, mkfs_function):
+        # clean the device
+        succ = BlockDev.fs_clean(self.loop_dev)
+
+        succ = mkfs_function(self.loop_dev, None)
+        self.assertTrue(succ)
+
+        # repair (expected to succeed)
+        succ = BlockDev.fs_repair(self.loop_dev)
+        self.assertTrue(succ)
+
+    def test_ext4_generic_repair(self):
+        """Test generic repair function with an ext4 file system"""
+        self._test_generic_repair(mkfs_function=BlockDev.fs_ext4_mkfs)
+
+    def test_xfs_generic_repair(self):
+        """Test generic repair function with an xfs file system"""
+        self._test_generic_repair(mkfs_function=BlockDev.fs_xfs_mkfs)
+
+class GenericSetLabel(FSTestCase):
+    def _test_generic_set_label(self, mkfs_function):
+        # clean the device
+        succ = BlockDev.fs_clean(self.loop_dev)
+
+        succ = mkfs_function(self.loop_dev, None)
+        self.assertTrue(succ)
+
+        # set label (expected to succeed)
+        succ = BlockDev.fs_set_label(self.loop_dev, "new_label")
+        self.assertTrue(succ)
+
+    def test_ext4_generic_set_label(self):
+        """Test generic set_label function with an ext4 file system"""
+        self._test_generic_set_label(mkfs_function=BlockDev.fs_ext4_mkfs)
+
+    def test_xfs_generic_set_label(self):
+        """Test generic set_label function with a xfs file system"""
+        self._test_generic_set_label(mkfs_function=BlockDev.fs_xfs_mkfs)
+
+class GenericResize(FSTestCase):
+    def _test_generic_resize(self, mkfs_function, fs_info_func=None, info_size_func=None):
+        # clean the device
+        succ = BlockDev.fs_clean(self.loop_dev)
+
+        succ = mkfs_function(self.loop_dev, None)
+        self.assertTrue(succ)
+
+        if info_size_func is not None and fs_info_func is not None:
+            size = info_size_func(fs_info_func(self.loop_dev))
+
         # shrink
         succ = BlockDev.fs_resize(self.loop_dev, 80 * 1024**2)
         self.assertTrue(succ)
+        if info_size_func is not None and fs_info_func is not None:
+            new_size = info_size_func(fs_info_func(self.loop_dev))
+            # do not check the size 100% precisely there may differences due to FS block size, etc.
+            self.assertEqual(new_size, 80 * 1024**2)
 
         # resize to maximum size
         succ = BlockDev.fs_resize(self.loop_dev, 0)
         self.assertTrue(succ)
+        if info_size_func is not None and fs_info_func is not None:
+            new_size = info_size_func(fs_info_func(self.loop_dev))
+            # should be back to original size
+            self.assertEqual(new_size, size)
 
     def test_ext2_generic_resize(self):
         """Test generic resize function with an ext2 file system"""
-        self._test_generic_resize(mkfs_function=BlockDev.fs_ext2_mkfs)
+        self._test_generic_resize(mkfs_function=BlockDev.fs_ext2_mkfs,
+                                  fs_info_func=BlockDev.fs_ext2_get_info,
+                                  info_size_func=lambda fi: fi.block_size * fi.block_count)
 
     def test_ext3_check_generic_resize(self):
         """Test generic resize function with an ext3 file system"""
-        self._test_generic_resize(mkfs_function=BlockDev.fs_ext3_mkfs)
+        self._test_generic_resize(mkfs_function=BlockDev.fs_ext3_mkfs,
+                                  fs_info_func=BlockDev.fs_ext3_get_info,
+                                  info_size_func=lambda fi: fi.block_size * fi.block_count)
 
     def test_ext4_generic_resize(self):
         """Test generic resize function with an ext4 file system"""
-        self._test_generic_resize(mkfs_function=BlockDev.fs_ext4_mkfs)
+        self._test_generic_resize(mkfs_function=BlockDev.fs_ext4_mkfs,
+                                  fs_info_func=BlockDev.fs_ext4_get_info,
+                                  info_size_func=lambda fi: fi.block_size * fi.block_count)
 
     @utils.skip_on("fedora", "27", reason="VFAT resize (detection after resize) is broken on rawhide")
     def test_vfat_generic_resize(self):
         """Test generic resize function with a vfat file system"""
         self._test_generic_resize(mkfs_function=BlockDev.fs_vfat_mkfs)
 
+    def _destroy_lvm(self):
+        run("vgremove --yes libbd_fs_tests &>/dev/null")
+        run("pvremove --yes %s &>/dev/null" % self.loop_dev)
+
     def test_xfs_generic_resize(self):
         """Test generic resize function with an xfs file system"""
 
+        run("pvcreate -ff -y %s &>/dev/null" % self.loop_dev)
+        run("vgcreate -s10M libbd_fs_tests %s &>/dev/null" % self.loop_dev)
+        run("lvcreate -n xfs_test -L50M libbd_fs_tests &>/dev/null")
+        self.addCleanup(self._destroy_lvm)
+
+        lv = "/dev/libbd_fs_tests/xfs_test"
+
         # clean the device
-        succ = BlockDev.fs_clean(self.loop_dev)
+        succ = BlockDev.fs_clean(lv)
 
-        succ = BlockDev.fs_xfs_mkfs(self.loop_dev, None)
+        succ = BlockDev.fs_xfs_mkfs(lv, None)
         self.assertTrue(succ)
 
-        # shrink without mounting the device (fs_resize should mount it)
-        succ = BlockDev.fs_resize(self.loop_dev, 0)
+        with mounted(lv, self.mount_dir):
+            fi = BlockDev.fs_xfs_get_info(lv)
+        self.assertTrue(fi)
+        self.assertEqual(fi.block_size * fi.block_count, 50 * 1024**2)
+
+        # no change, nothing should happen
+        with mounted(lv, self.mount_dir):
+            succ = BlockDev.fs_resize(lv, 0)
         self.assertTrue(succ)
 
-        # resize to maximum with mounting the device (fs_resize should use
-        # the existing mountpoint)
-        with mounted(self.loop_dev, self.mount_dir):
-            succ = BlockDev.fs_resize(self.loop_dev, 0)
+        with mounted(lv, self.mount_dir):
+            fi = BlockDev.fs_xfs_get_info(lv)
+        self.assertTrue(fi)
+        self.assertEqual(fi.block_size * fi.block_count, 50 * 1024**2)
+
+        # (still) impossible to shrink an XFS file system
+        with mounted(lv, self.mount_dir):
+            with self.assertRaises(GLib.GError):
+                succ = BlockDev.fs_resize(lv, 40 * 1024**2)
+
+        run("lvresize -L70M libbd_fs_tests/xfs_test &>/dev/null")
+        # should grow
+        with mounted(lv, self.mount_dir):
+            succ = BlockDev.fs_resize(lv, 0)
         self.assertTrue(succ)
+        with mounted(lv, self.mount_dir):
+            fi = BlockDev.fs_xfs_get_info(lv)
+        self.assertTrue(fi)
+        self.assertEqual(fi.block_size * fi.block_count, 70 * 1024**2)
+
+        run("lvresize -L90M libbd_fs_tests/xfs_test &>/dev/null")
+        # should grow just to 80 MiB
+        with mounted(lv, self.mount_dir):
+            succ = BlockDev.fs_resize(lv, 80 * 1024**2)
+        self.assertTrue(succ)
+        with mounted(lv, self.mount_dir):
+            fi = BlockDev.fs_xfs_get_info(lv)
+        self.assertTrue(fi)
+        self.assertEqual(fi.block_size * fi.block_count, 80 * 1024**2)
+
+        # should grow to 90 MiB
+        with mounted(lv, self.mount_dir):
+            succ = BlockDev.fs_resize(lv, 0)
+        self.assertTrue(succ)
+        with mounted(lv, self.mount_dir):
+            fi = BlockDev.fs_xfs_get_info(lv)
+        self.assertTrue(fi)
+        self.assertEqual(fi.block_size * fi.block_count, 90 * 1024**2)
